//---------------------------------------------------------------------- 
//----------------------------------------------------------------------
//
// AMR Driver for the Mike's CellVertex Application
// Buckley Leverette Equations....
//
//----------------------------------------------------------------------
//---------------------------------------------------------------------- 

//---------------------------------------------------------------------- 
// DAGH includes 
//---------------------------------------------------------------------- 
#include "DAGH.h"
#include "DAGHCluster.h"
#include "DAGHIO.h"
//---------------------------------------------------------------------- 

//---------------------------------------------------------------------- 
// Application Includes
//---------------------------------------------------------------------- 
#include "mikecv.h"
#include "mikecvfortran2d.h"

//---------------------------------------------------------------------- 
// Flags
//---------------------------------------------------------------------- 
#define VizServer 0 	// Who will write to stdout
#define DAGH_IO 0   	// Are we using the DAGH IO server

//---------------------------------------------------------------------- 
// AMR parameters
//---------------------------------------------------------------------- 
INTEGER MaxLev = 1;		// Maximum level of refinement
INTEGER BufferWidth = 1;	// Buffer width used by clusterer
INTEGER BlockWidth = 1;		// Minimum granularity used by clusterer
DOUBLE MinEfficiency = 0.7;	// Minumum efficiency of clustering
GFTYPE Thresh = 0.0;		// Truncation error threshold for regriding
INTEGER RegridEvery = 4;	// How oftern do I regrid ?
INTEGER NumIters = 0;		// Number of timesteps on the base grid

//---------------------------------------------------------------------- 
// IO control parameters
//---------------------------------------------------------------------- 
INTEGER OutEvery = 0;

void main(INTEGER argc, CHARACTER *argv[]) 
{

//---------------------------------------------------------------
// If I using MPI initialize it...
//---------------------------------------------------------------
#ifdef DAGH_NO_MPI
#else
   cout << "Initializing MPI\n";
   MPI_Init( &argc, &argv ); // Initialize MPI
#endif

//---------------------------------------------------------------
// If I using DAGHIO support initialize it...
//---------------------------------------------------------------
#ifdef DAGH_IO 
  cout << "Initializing DAGHIO\n";
  DAGHIOInit();
#endif

//---------------------------------------------------------------
// Local variable declarations
//---------------------------------------------------------------
   INTEGER  nx, ny;

   DOUBLE  h, dt;                         // Grid spacings
   DOUBLE xmin, xmax, ymin, ymax;         // Grid extent
   DOUBLE cfl;  			  // Initial data parameters

//---------------------------------------------------------------
// Read grid properties and initial data parameters from file
//---------------------------------------------------------------
{
   INTEGER niters = 0;
   INTEGER ml, regride, buffw, blkw, outevery;
   GFTYPE thresh = 0;
   DOUBLE mineff = 0;
   INTEGER d = 0;

   cout << "****Reading in parameters****" << endl << flush;
   f_readinput(&nx, &ny, &xmin, &xmax, &ymin, &ymax, &cfl,
               &niters,    // Number of iterations
               &ml,        // Max levels
               &thresh,    // Threshold
               &regride,   // Regrid every
               &mineff,    // Min efficiency
               &buffw,     // Buffer width
               &blkw,      // Min block width
               &outevery); // Output every

   MaxLev = ml;
   RegridEvery = regride;
   BufferWidth = buffw;
   BlockWidth = blkw;
   MinEfficiency = mineff;
   Thresh = thresh;
   OutEvery = outevery;
   NumIters = niters;
}

//---------------------------------------------------------------
// Calculate h and dt from inputs
//---------------------------------------------------------------
   h  = (xmax - xmin) / (DOUBLE)(nx-1);
   dt = h * cfl;

//---------------------------------------------------------------
// Write out properties
//---------------------------------------------------------------
   cout << "******************************************************************" << endl << flush;
   cout << "****Grid Properties****" << endl << flush;
   cout << "nx = " << nx << " ny =" << ny << endl << flush;
   cout << "xmin = " << xmin << " xmax = " << xmax << endl << flush;
   cout << "ymin = " << ymin << " ymax = " << ymax << endl << flush;
   cout << "h = " << h << " dt = " << dt << endl << flush;

   cout << "****AMR Specifications****" << endl << flush;
   cout << "max lev = " << MaxLev <<  flush;
   cout << "threshold = " << Thresh << endl << flush;
   cout << "regrid every = " << RegridEvery << endl << flush;
   cout << "buffer width = " << BufferWidth << endl << flush;
   cout << "block width = " << BlockWidth << endl << flush;
   cout << "min efficiency = " << MinEfficiency << endl << flush;
   cout << "******************************************************************" << endl << flush;

//---------------------------------------------------------------
// Setup grid hierarchy
//---------------------------------------------------------------
   INTEGER  shape[DIM];
   DOUBLE  bb[2*DIM];

   shape[0] = nx-1;
   shape[1] = ny-1;

   bb[0] = xmin ;
   bb[1] = xmax ;
   bb[2] = ymin ;
   bb[3] = ymax ;

   GridHierarchy GH(DIM,DAGHVertexCentered,MaxLev);
   SetBaseGrid (GH,bb,shape);
   SetRefineFactor(GH,2);
   SetBoundaryWidth(GH,1);
   SetBoundaryType(GH,DAGHNoBoundary);

#ifdef DAGH_IO
   DAGHIOType(GH,DAGHIO_HDF_NCSA);
#endif

  ComposeHierarchy(GH); 

//---------------------------------------------------------------
// Processor information
//---------------------------------------------------------------
  INTEGER me = MY_PROC(GH);
  INTEGER num = NUM_PROC(GH);

  cout << "I AM PROCESSOR " << me << " OF " << num-1 << endl << flush;

BEGIN_COMPUTE
//---------------------------------------------------------------
//  Declare grid functions
//---------------------------------------------------------------
   INTEGER  t_sten = 0;
   INTEGER  s_sten = 1;

   /* VertexCellCentered GF's */
   GridFunction(DIM)<GFTYPE> u("u", t_sten, s_sten, GH, DAGHCommFaceOnly, DAGHHasShadow);
   GridFunction(DIM)<GFTYPE> temp("temp", t_sten, s_sten, GH, DAGHCommFaceOnly, DAGHHasShadow);

   GridFunction(DIM)<GFTYPE> fu("fu", t_sten, s_sten, GH, DAGHNoComm, DAGHHasShadow);
   GridFunction(DIM)<GFTYPE> gu("gu", t_sten, s_sten, GH, DAGHNoComm, DAGHHasShadow);
   GridFunction(DIM)<GFTYPE> hu("hu", t_sten, s_sten, GH, DAGHNoComm, DAGHHasShadow);

   foreachGF(GH,GF,DIM,GFTYPE)
     SetProlongFunction(GF, (void *) &f_prolong_amr);
     SetRestrictFunction(GF, (void *) &f_restrict_amr);
   end_foreachGF

//---------------------------------------------------------------
// Local variables needed for evolution
//---------------------------------------------------------------
   GFTYPE unorm;

//---------------------------------------------------------------
// Initialize Level = 0 (Base Grid)
//---------------------------------------------------------------
  INTEGER Level = 0;
  INTEGER l = Level;
  INTEGER t = CurrentTime(GH,Level,DAGH_Main);
  INTEGER idt = TimeStep(GH,Level,DAGH_Main);

//---------------------------------------------------------------
// Set up initial data
//---------------------------------------------------------------
   if (me == VizServer)
     cout << "****Setting up Initial data on DAGH_Main ****" << endl << flush;

   forall(u, t, l, c)
     f_initial_data(FBA(u(t,l,c,DAGH_Main)), FDA(u(t,l,c,DAGH_Main)));
   end_forall

   if (me == VizServer)
     cout << "****Setting up Initial data on DAGH_Shadow ****" << endl << flush;

   forall(u, t, l, c)
     f_initial_data(FBA(u(t,l,c,DAGH_Shadow)), FDA(u(t,l,c,DAGH_Shadow)));
   end_forall

//---------------------------------------------------------------
// Dump Initial Data on Main
//---------------------------------------------------------------
#ifdef DAGH_IO
      for(INTEGER curlev = 0; curlev <= FineLevel(GH); curlev++) {
        if (me == VizServer) cout << "DAGHIO: Dumping Level" << curlev << endl << flush;
        INTEGER myt = CurrentTime(GH,curlev,DAGH_Main);
        const INTEGER saveTime = CurrentTime(GH,curlev);
	GH.setCurrentTime(CurrentTime(GH,curlev)/TimeStep(GH,0),curlev);
	Write(u,CurrentTime(GH,curlev),curlev,DAGH_Main,DAGH_Double);
        GH.setCurrentTime(saveTime,curlev);
      }
#endif

//---------------------------------------------------------------
// Begin Evolution
//---------------------------------------------------------------
  INTEGER currentiter = 0;
  if (me == VizServer) cout << "Starting main loop" << endl << flush;

  for (currentiter=1;currentiter<=NumIters;currentiter++) {

      // Carry out evolution steps on all grids on all levels
      // At this stage only timelevels n and n-1 are defined.

      bno_RecursiveIntegrate(GH, Level, u, fu, gu, hu, temp, dt, h, cfl);

//---------------------------------------------------------------
// Dump Info
//---------------------------------------------------------------
#ifdef DAGH_IO
      if (StepsTaken(GH,Level,DAGH_Main) % OutEvery == 0) {
        for(INTEGER curlev = 0; curlev <= FineLevel(GH); curlev++) {
          if (me == VizServer) cout << "DAGHIO: Dumping Level" << curlev << endl << flush;
          INTEGER myt = CurrentTime(GH,curlev,DAGH_Main);
	  const INTEGER saveTime = CurrentTime(GH,curlev);
	  GH.setCurrentTime(CurrentTime(GH,curlev)/TimeStep(GH,0),curlev);
	  Write(u,CurrentTime(GH,curlev),curlev,DAGH_Main,DAGH_Double);
          GH.setCurrentTime(saveTime,curlev);
      }}
#endif

  } // End loop over number of iterations

#ifdef DAGH_IO
  DAGHIOEnd(GH);
#endif
END_COMPUTE

#ifdef DAGH_IO
  DAGHIO_HDF_NCSA_Finalize();
#endif

#ifdef DAGH_NO_MPI
#else
 cout << "Finalizing MPI\n";
  DAGHMPI_Finalize();
#endif

} // End main

//----------------------------------------------------------------------
// 
// Recursive Integrate 
//                        
//----------------------------------------------------------------------
void bno_RecursiveIntegrate(
                GridHierarchy &GH,
                INTEGER const Level,
                GridFunction(DIM)<GFTYPE> &u,
                GridFunction(DIM)<GFTYPE> &fu,
                GridFunction(DIM)<GFTYPE> &gu,
                GridFunction(DIM)<GFTYPE> &hu,
                GridFunction(DIM)<GFTYPE> &temp,
                DOUBLE &dt, DOUBLE &h, DOUBLE const &cfl)
{
//---------------------------------------------------------------
// Processor Info
//---------------------------------------------------------------
  INTEGER me = MY_PROC(GH);
  INTEGER num = NUM_PROC(GH);

//---------------------------------------------------------------
// Select number of iterations to run based on current grid level
//---------------------------------------------------------------
  INTEGER NoIterations = 0;
  if (Level==0)
      NoIterations = 1;
  else
     NoIterations = RefineFactor(GH);

//---------------------------------------------------------------
// Take recursive steps..
//---------------------------------------------------------------
  for (INTEGER i=0;i<NoIterations;i++) {

    INTEGER l = Level;
    INTEGER t = CurrentTime(GH,Level,DAGH_Main);
    INTEGER idt = TimeStep(GH,Level,DAGH_Main);

//---------------------------------------------------------------
// Is it regridding time?
//---------------------------------------------------------------
    if (Level < MaxLevel(GH) &&
        StepsTaken(GH,Level,DAGH_Main)%RegridEvery == 0 &&
        StepsTaken(GH,Level,DAGH_Main) != 0) {

        if (me == VizServer)
          cout << "********* Regriding at Level" << Level <<"*********" << endl << flush ;

       // Truncation error estimation using DAGH_Main
       wave_trunc_est(GH,t,l,u,temp);

       // Regrid using truncation error estimate
       bno_RegridSystemAbove(GH,temp,Level,Thresh);

    } // End if statement Level...

//---------------------------------------------------------------
// Take a step on the DAGH_Main hierarchy 
//---------------------------------------------------------------
    DOUBLE dagh_h = DeltaX(GH,DAGH_X,l,DAGH_Main);
    DOUBLE dagh_dt = dagh_h * cfl;

    if (me == VizServer) 
      cout << "Integrate on DAGH_Main - Level " << l << endl << flush;
    forall(u, t, l, c)
       f_update(FBA(u(t,l,c,DAGH_Main)),
                FDA(u(t,l,c,DAGH_Main)), FDA(fu(t,l,c,DAGH_Main)),
                FDA(gu(t,l,c,DAGH_Main)), FDA(hu(t,l,c,DAGH_Main)),
                &dagh_dt, &dagh_h);
    end_forall

    if (me == VizServer) 
      cout << "Updating boundaries on DAGH_Main - Level " << l << endl << flush ; 
    BoundaryUpdate(u,t,l,DAGH_Main);

    forall(u, t, l, c)
       f_update_boundary(FBA(u(t,l,c,DAGH_Main)),
                FDA(u(t,l,c,DAGH_Main)), FDA(fu(t,l,c,DAGH_Main)),
                FDA(gu(t,l,c,DAGH_Main)), FDA(hu(t,l,c,DAGH_Main)),
                &dagh_dt, &dagh_h);
    end_forall

    Sync(u, t, l, DAGH_Main);
    
//---------------------------------------------------------------
// Take a step on the DAGH_Shadow hierarchy 
//---------------------------------------------------------------
    if (StepsTaken(GH,Level,DAGH_Main) % RefineFactor(GH) == 0){

       DOUBLE dagh_h = DeltaX(GH,DAGH_X,l,DAGH_Shadow);
       DOUBLE dagh_dt = dagh_h * cfl;

       // First update DAGH_Shadow from DAGH_Main
       forall(u, t, l, c)
         u(t,l,c,DAGH_Shadow) = u(t,l,c,DAGH_Main);
         fu(t,l,c,DAGH_Shadow) = fu(t,l,c,DAGH_Main);
         gu(t,l,c,DAGH_Shadow) = gu(t,l,c,DAGH_Main);
         hu(t,l,c,DAGH_Shadow) = hu(t,l,c,DAGH_Main);
       end_forall
 
       Sync(u, t, l, DAGH_Shadow);

       if (me == VizServer) 
         cout << "Integrate on DAGH_Shadow" << " - Level " << l << endl << flush;

       // flow on DAGH_Shadow
       forall(u, t, l, c)
          f_update(FBA(u(t,l,c,DAGH_Shadow)),
                FDA(u(t,l,c,DAGH_Shadow)), FDA(fu(t,l,c,DAGH_Shadow)),
                FDA(gu(t,l,c,DAGH_Shadow)), FDA(hu(t,l,c,DAGH_Shadow)),
                &dagh_dt, &dagh_h);
       end_forall

       if(me == VizServer) cout << "Updating boundaries on DAGH_Shadow" 
                                << " - Level " << l << endl << flush ; 
       forall(u, t, l, c)
          f_update_boundary(FBA(u(t,l,c,DAGH_Shadow)),
                FDA(u(t,l,c,DAGH_Shadow)), FDA(fu(t,l,c,DAGH_Shadow)),
                FDA(gu(t,l,c,DAGH_Shadow)), FDA(hu(t,l,c,DAGH_Shadow)),
                &dagh_dt, &dagh_h);
       end_forall

       Sync(u, t, l, DAGH_Shadow);

    } // Ends if statement for DAGH_Shadow
    
//---------------------------------------------------------------
// If we are not at the finest level then go to next level
//---------------------------------------------------------------
    if (Level < FineLevel(GH)) {
      if(me == VizServer) cout << "######################Integration of Level " 
                 << Level+1 << endl << flush ;

       // Recursive Integration of next level in hierarchy
       bno_RecursiveIntegrate( GH, Level+1, u, fu, gu, hu, temp, dt, h, cfl );

       if(me == VizServer) cout << "######################Done with Integration of Level " 
                 << Level+1 << endl << flush ;

    } // End if (Level < FineLevel(GH))

//---------------------------------------------------------------
// Increment time step on current level
//---------------------------------------------------------------
    IncrCurrentTime(GH,Level,DAGH_Main);

//---------------------------------------------------------------
// Restriction on DAGH_Main
//---------------------------------------------------------------
    if (Level < FineLevel(GH)) {
      /* a sanity check */
      assert (CurrentTime(GH,Level,DAGH_Main) == CurrentTime(GH,Level+1,DAGH_Main)); 

      INTEGER t = CurrentTime(GH,Level,DAGH_Main);
      INTEGER myargc = 0; GFTYPE myargs[1];
      Restrict(u, t, Level+1, t, Level, myargs, myargc, DAGH_Main);
      Sync(u, t, Level, DAGH_Main);
    } // end if statement for level

    cout << "Steps taken = " << StepsTaken(GH,Level,DAGH_Main)
         << " at Level = " << Level << endl << flush;

  } // End For loop

} // End bnoRecursiveIntegrate

//---------------------------------------------------------------
// Regrid GridHierarchy at Level
//---------------------------------------------------------------
void bno_RegridSystemAbove(GridHierarchy &GH,
                           GridFunction(DIM)<GFTYPE>& u,
                           INTEGER const Level, GFTYPE const thresh)
{
  INTEGER me = MY_PROC(GH);
  INTEGER num = NUM_PROC(GH);

//---------------------------------------------------------------
// Compute levels to regrid at - from the finest down to Level
//---------------------------------------------------------------
  INTEGER flev = FineLevel(GH);
  INTEGER l = 0;
  if (flev == 0) l = 0;
  else if (flev == MaxLev-1) l = flev-1;
  else l = flev;

//---------------------------------------------------------------
// Normalize the threshhold (I'm not sure why Mike does this)
//---------------------------------------------------------------
  GFTYPE maxte = 0;
  for (INTEGER ll=l ;ll>=Level;ll--) {
    INTEGER t = CurrentTime(GH,Level,DAGH_Main);
    GFTYPE tmpmax = MaxVal(u,t,ll,DAGH_Main);
    maxte = (maxte > tmpmax) ? maxte : tmpmax;
  }

//---------------------------------------------------------------
// Cluster at refine
//---------------------------------------------------------------
  BBoxList bblist;

  for (;l>=Level;l--) {
    INTEGER t = CurrentTime(GH,l,DAGH_Main);
    INTEGER idt = TimeStep(GH,l,DAGH_Main);

    DAGHCluster2d(u, t, l,
                  BlockWidth, BufferWidth,
                  MinEfficiency, maxte*thresh,
                  bblist, bblist, DAGH_Main);

    Refine(GH,bblist,l);

  } // Ends for loop over levels

//---------------------------------------------------------------
// Recompose the GridHierarchy
//---------------------------------------------------------------
  RecomposeHierarchy(GH);

} // bno_RegridSystemAbove
